package com.qihoo.qsql.api;

import com.qihoo.qsql.exception.UnsupportedApiException;
import com.qihoo.qsql.metadata.MetadataPostman;
import com.qihoo.qsql.exec.JdbcPipeline;
import com.qihoo.qsql.plan.QueryTables;
import com.qihoo.qsql.utils.SqlUtil;
import java.sql.Array;
import java.sql.Blob;
import java.sql.CallableStatement;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.NClob;
import java.sql.PreparedStatement;
import java.sql.SQLClientInfoException;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.SQLXML;
import java.sql.Savepoint;
import java.sql.Statement;
import java.sql.Struct;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.Executor;

/**
 * JDBC connection implementation.
 */
public class AutomaticConnection implements Connection {

    private static final String VALUES_SCHEMA = "inline:"
        + "{\n"
        + "  version: '1.0',\n"
        + "  defaultSchema: 'SALES',\n"
        + "  schemas: [\n"
        + "    {\n"
        + "      name: 'SALES',\n"
        + "      type: 'custom',\n"
        + "      factory: 'org.apache.calcite.adapter.csv.CsvSchemaFactory',\n"
        + "      operand: {\n"
        + "        directory: 'sales'\n"
        + "      }\n"
        + "    }\n"
        + "  ]\n"
        + "}";
    private Connection connection;
    private String schemaPath = "";
    private Set<String> tableNames = new HashSet<>();
    private Connection simpleConnection;

    public AutomaticConnection(String path) throws SQLException {
        this();
        this.schemaPath = path;
    }

    /**
     * Automatic Connection based on json config.
     */
    public AutomaticConnection() throws SQLException {
        Properties info = new Properties();
        info.put("model", VALUES_SCHEMA);
        this.simpleConnection = DriverManager.getConnection("jdbc:calcite:", info);
    }

    @Override
    public PreparedStatement prepareStatement(String sql) throws SQLException {
        QueryTables tables = SqlUtil.parseTableName(sql);
        if (tables.isDml()) {
            throw new RuntimeException("Unsupported DML in JDBC model");
        }

        List<String> names = tables.tableNames;
        if (names.isEmpty()) {
            return simpleConnection.prepareStatement(sql);
        }

        //caution: tableName has changed to dbName.tableName
        if (names.stream().anyMatch(name -> ! tableNames.contains(name))) {
            tableNames.clear();
            tableNames.addAll(names);
        } else {
            if (connection != null) {
                return connection.prepareStatement(sql);
            }
        }

        if (schemaPath.isEmpty()) {
            this.connection = JdbcPipeline.createSpecificConnection(
                MetadataPostman.getAssembledSchema(new ArrayList<>(tableNames)));
        } else {
            this.connection =
                JdbcPipeline.createSpecificConnection(schemaPath, new ArrayList<>(tableNames));
        }

        return connection.prepareStatement(sql);
    }

    @Override
    public PreparedStatement prepareStatement(String sql,
        int resultSetType,
        int resultSetConcurrency) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public PreparedStatement prepareStatement(String sql,
        int resultSetType,
        int resultSetConcurrency,
        int resultSetHoldability) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
        throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public PreparedStatement prepareStatement(String sql, int[] columnIndexes) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public PreparedStatement prepareStatement(String sql, String[] columnNames)
        throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public CallableStatement prepareCall(String sql) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public CallableStatement prepareCall(String sql,
        int resultSetType,
        int resultSetConcurrency) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public CallableStatement prepareCall(String sql,
        int resultSetType,
        int resultSetConcurrency,
        int resultSetHoldability) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public String nativeSQL(String sql) throws SQLException {
        return connection.nativeSQL(sql);
    }

    @Override
    public boolean getAutoCommit() throws SQLException {
        return connection.getAutoCommit();
    }

    @Override
    public void setAutoCommit(boolean autoCommit) throws SQLException {
        connection.setAutoCommit(autoCommit);
    }

    @Override
    public void commit() throws SQLException {
        connection.commit();
    }

    @Override
    public void rollback() throws SQLException {
        connection.rollback();
    }

    @Override
    public void close() throws SQLException {
        connection.close();
    }

    @Override
    public boolean isClosed() throws SQLException {
        return connection.isClosed();
    }

    @Override
    public DatabaseMetaData getMetaData() throws SQLException {
        return connection.getMetaData();
    }

    @Override
    public boolean isReadOnly() throws SQLException {
        return connection.isReadOnly();
    }

    @Override
    public void setReadOnly(boolean readOnly) throws SQLException {
        connection.setReadOnly(readOnly);
    }

    @Override
    public String getCatalog() throws SQLException {
        return connection.getCatalog();
    }

    @Override
    public void setCatalog(String catalog) throws SQLException {
        connection.setCatalog(catalog);
    }

    @Override
    public int getTransactionIsolation() throws SQLException {
        return connection.getTransactionIsolation();
    }

    @Override
    public void setTransactionIsolation(int level) throws SQLException {
        connection.setTransactionIsolation(level);
    }

    @Override
    public SQLWarning getWarnings() throws SQLException {
        return connection.getWarnings();
    }

    @Override
    public void clearWarnings() throws SQLException {
        connection.clearWarnings();
    }

    @Override
    public Statement createStatement() throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public Statement createStatement(int resultSetType,
        int resultSetConcurrency,
        int resultSetHoldability) throws SQLException {
        throw new UnsupportedApiException("");
    }

    @Override
    public Statement createStatement(int resultSetType,
        int resultSetConcurrency) throws SQLException {
        throw new UnsupportedApiException("");
    }


    @Override
    public Map<String, Class<?>> getTypeMap() throws SQLException {
        return connection.getTypeMap();
    }

    @Override
    public void setTypeMap(Map<String, Class<?>> map) throws SQLException {
        connection.setTypeMap(map);
    }

    @Override
    public int getHoldability() throws SQLException {
        return connection.getHoldability();
    }

    @Override
    public void setHoldability(int holdability) throws SQLException {
        connection.setHoldability(holdability);
    }

    @Override
    public Savepoint setSavepoint() throws SQLException {
        return connection.setSavepoint();
    }

    @Override
    public Savepoint setSavepoint(String name) throws SQLException {
        return connection.setSavepoint(name);
    }

    @Override
    public void rollback(Savepoint savepoint) throws SQLException {
        connection.rollback(savepoint);
    }

    @Override
    public void releaseSavepoint(Savepoint savepoint) throws SQLException {
        connection.releaseSavepoint(savepoint);
    }


    @Override
    public Clob createClob() throws SQLException {
        return connection.createClob();
    }

    @Override
    public Blob createBlob() throws SQLException {
        return connection.createBlob();
    }

    @Override
    public NClob createNClob() throws SQLException {
        return connection.createNClob();
    }

    @Override
    public SQLXML createSQLXML() throws SQLException {
        return connection.createSQLXML();
    }

    @Override
    public boolean isValid(int timeout) throws SQLException {
        return connection.isValid(timeout);
    }

    @Override
    public void setClientInfo(String name, String value) throws SQLClientInfoException {
        connection.setClientInfo(name, value);
    }

    @Override
    public String getClientInfo(String name) throws SQLException {
        return connection.getClientInfo(name);
    }

    @Override
    public Properties getClientInfo() throws SQLException {
        return connection.getClientInfo();
    }

    @Override
    public void setClientInfo(Properties properties) throws SQLClientInfoException {
        connection.setClientInfo(properties);
    }

    @Override
    public Array createArrayOf(String typeName, Object[] elements) throws SQLException {
        return connection.createArrayOf(typeName, elements);
    }

    @Override
    public Struct createStruct(String typeName, Object[] attributes) throws SQLException {
        return connection.createStruct(typeName, attributes);
    }

    @Override
    public String getSchema() throws SQLException {
        return connection.getSchema();
    }

    @Override
    public void setSchema(String schema) throws SQLException {
        connection.setSchema(schema);
    }

    @Override
    public void abort(Executor executor) throws SQLException {
        connection.abort(executor);
    }

    @Override
    public void setNetworkTimeout(Executor executor, int milliseconds) throws SQLException {
        connection.setNetworkTimeout(executor, milliseconds);
    }

    @Override
    public int getNetworkTimeout() throws SQLException {
        return connection.getNetworkTimeout();
    }

    @Override
    public <T> T unwrap(Class<T> iface) throws SQLException {
        return connection.unwrap(iface);
    }

    @Override
    public boolean isWrapperFor(Class<?> iface) throws SQLException {
        return connection.isWrapperFor(iface);
    }

}
